@@ -1087,9 +1087,14 @@ typedef struct nghttp3_qpack_encoder nghttp3_qpack_encoder;
  * @function
  *
  * `nghttp3_qpack_encoder_new` initializes QPACK encoder.  |pencoder|
- * must be non-NULL pointer.  |mem| is a memory allocator.  This
- * function allocates memory for :type:`nghttp3_qpack_encoder` itself
- * and assigns its pointer to |*pencoder| if it succeeds.
+ * must be non-NULL pointer.  |hard_max_dtable_capacity| is the upper
+ * bound of the dynamic table capacity.  |mem| is a memory allocator.
+ * This function allocates memory for :type:`nghttp3_qpack_encoder`
+ * itself and assigns its pointer to |*pencoder| if it succeeds.
+ *
+ * The maximum dynamic table capacity is still 0.  In order to change
+ * the maximum dynamic table capacity, call
+ * `nghttp3_qpack_encoder_set_max_dtable_capacity`.
  *
  * This function returns 0 if it succeeds, or one of the following
  * negative error codes:
@@ -1098,6 +1103,7 @@ typedef struct nghttp3_qpack_encoder nghttp3_qpack_encoder;
  *     Out of memory.
  */
 NGHTTP3_EXTERN int nghttp3_qpack_encoder_new(nghttp3_qpack_encoder **pencoder,
+                                             size_t hard_max_dtable_capacity,
                                              const nghttp3_mem *mem);
 
 /**
@@ -1166,41 +1172,14 @@ NGHTTP3_EXTERN nghttp3_ssize nghttp3_qpack_encoder_read_decoder(
  * @function
  *
  * `nghttp3_qpack_encoder_set_max_dtable_capacity` sets max dynamic
- * table size to |max_dtable_capacity|.  |max_dtable_capacity| must be
- * equal to or smaller than the value passed to
- * `nghttp3_qpack_encoder_set_max_dtable_capacity`.
- *
- * This function returns 0 if it succeeds, or one of the
- * following negative error codes:
- *
- * :macro:`NGHTTP3_ERR_INVALID_ARGUMENT`
- *     |max_dtable_capacity| exceeds the hard limit that decoder
- *     specifies.
+ * table capacity to |max_dtable_capacity|.  If |max_dtable_capacity| is
+ * larger than ``hard_max_dtable_capacity`` parameter of
+ * `nghttp3_qpack_encoder_new`, it is truncated to the latter.
  */
-NGHTTP3_EXTERN int
+NGHTTP3_EXTERN void
 nghttp3_qpack_encoder_set_max_dtable_capacity(nghttp3_qpack_encoder *encoder,
                                               size_t max_dtable_capacity);
 
-/**
- * @function
- *
- * `nghttp3_qpack_encoder_set_hard_max_dtable_capacity` sets hard
- * maximum dynamic table size to |hard_max_dtable_capacity|.  This is
- * the real maximum capacity of dynamic table that encoder enforces.
- * This value is used to validate the passed value by
- * `nghttp3_qpack_encoder_set_max_dtable_capacity`.
- *
- * In HTTP/3, QPACK decoder side of the connection tells the maximum
- * size in SETTINGS_QPACK_MAX_TABLE_CAPACITY.  It is dangerous to pass
- * the value as is to this function because the decoder can set the
- * large table size which may increase the memory consumption of
- * encoder.  Therefore, it should be the minimum of the value of
- * SETTINGS_QPACK_MAX_TABLE_CAPACITY and the amount of memory that
- * encoder can afford.
- */
-NGHTTP3_EXTERN void nghttp3_qpack_encoder_set_hard_max_dtable_capacity(
-    nghttp3_qpack_encoder *encoder, size_t hard_max_dtable_capacity);
-
 /**
  * @function
  *
@@ -1299,11 +1278,12 @@ typedef struct nghttp3_qpack_decoder nghttp3_qpack_decoder;
  * @function
  *
  * `nghttp3_qpack_decoder_new` initializes QPACK decoder.  |pdecoder|
- * must be non-NULL pointer.  |max_dtable_capacity| is the maximum
- * dynamic table size.  |max_blocked_streams| is the maximum number of
- * streams which can be blocked.  |mem| is a memory allocator.  This
- * function allocates memory for :type:`nghttp3_qpack_decoder` itself
- * and assigns its pointer to |*pdecoder| if it succeeds.
+ * must be non-NULL pointer.  |hard_max_dtable_capacity| is the upper
+ * bound of the dynamic table capacity.  |max_blocked_streams| is the
+ * maximum number of streams which can be blocked.  |mem| is a memory
+ * allocator.  This function allocates memory for
+ * :type:`nghttp3_qpack_decoder` itself and assigns its pointer to
+ * |*pdecoder| if it succeeds.
  *
  * This function returns 0 if it succeeds, or one of the following
  * negative error codes:
@@ -1312,7 +1292,7 @@ typedef struct nghttp3_qpack_decoder nghttp3_qpack_decoder;
  *     Out of memory.
  */
 NGHTTP3_EXTERN int nghttp3_qpack_decoder_new(nghttp3_qpack_decoder **pdecoder,
-                                             size_t max_dtable_capacity,
+                                             size_t hard_max_dtable_capacity,
                                              size_t max_blocked_streams,
                                              const nghttp3_mem *mem);
 
@@ -1479,11 +1459,21 @@ nghttp3_qpack_decoder_cancel_stream(nghttp3_qpack_decoder *decoder,
  * @function
  *
  * `nghttp3_qpack_decoder_set_max_dtable_capacity` sets
- * |max_dtable_capacity| as maximum dynamic table size.  Normally, the
- * maximum capacity is communicated in encoder stream.  This function
- * is provided for debugging and testing purpose.
+ * |max_dtable_capacity| as maximum dynamic table size.
+ * |max_dtable_capacity| must be equal to or smaller than
+ * ``hard_max_dtable_capacity`` parameter of
+ * `nghttp3_qpack_decoder_new`.  Normally, the maximum capacity is
+ * communicated in encoder stream.  This function is provided for
+ * debugging and testing purpose.
+ *
+ * This function returns 0 if it succeeds, or one of the
+ * following negative error codes:
+ *
+ * :macro:`NGHTTP3_ERR_INVALID_ARGUMENT`
+ *     |max_dtable_capacity| exceeds the upper bound of the dynamic
+ *     table capacity.
  */
-NGHTTP3_EXTERN void
+NGHTTP3_EXTERN int
 nghttp3_qpack_decoder_set_max_dtable_capacity(nghttp3_qpack_decoder *decoder,
                                               size_t max_dtable_capacity);
 
@@ -1708,13 +1698,15 @@ typedef int (*nghttp3_recv_header)(nghttp3_conn *conn, int64_t stream_id,
  * :type:`nghttp3_end_headers` is a callback function which is invoked
  * when an incoming header block has ended.
  *
+ * If the stream ends with this header block, |fin| is set to nonzero.
+ *
  * The implementation of this callback must return 0 if it succeeds.
  * Returning :macro:`NGHTTP3_ERR_CALLBACK_FAILURE` will return to the
  * caller immediately.  Any values other than 0 is treated as
  * :macro:`NGHTTP3_ERR_CALLBACK_FAILURE`.
  */
 typedef int (*nghttp3_end_headers)(nghttp3_conn *conn, int64_t stream_id,
-                                   void *conn_user_data,
+                                   int fin, void *conn_user_data,
                                    void *stream_user_data);
 
 /**
@@ -1900,10 +1892,19 @@ typedef struct nghttp3_settings {
    */
   uint64_t max_field_section_size;
   /**
-   * :member:`qpack_max_table_capacity` is the maximum size of QPACK
+   * :member:`qpack_max_dtable_capacity` is the maximum size of QPACK
    * dynamic table.
    */
-  size_t qpack_max_table_capacity;
+  size_t qpack_max_dtable_capacity;
+  /**
+   * :member:`qpack_encoder_max_dtable_capacity` is the upper bound of
+   * QPACK dynamic table capacity that the QPACK encoder is willing to
+   * use.  The effective maximum dynamic table capacity is the minimum
+   * of this field and the value of the received
+   * SETTINGS_QPACK_MAX_TABLE_CAPACITY.  If this field is set to 0,
+   * the encoder does not use the dynamic table.
+   */
+  size_t qpack_encoder_max_dtable_capacity;
   /**
    * :member:`qpack_blocked_streams` is the maximum number of streams
    * which can be blocked while they are being decoded.
@@ -1923,6 +1924,17 @@ typedef struct nghttp3_settings {
  *
  * `nghttp3_settings_default` fills |settings| with the default
  * values.
+ *
+ * - :member:`max_field_section_size
+ *   <nghttp3_settings.max_field_section_size>` = :expr:`((1ull << 62) - 1)`
+ * - :member:`qpack_max_dtable_capacity
+ *   <nghttp3_settings.qpack_max_dtable_capacity>` = 0
+ * - :member:`qpack_encoder_max_dtable_capacity
+ *   <nghttp3_settings.qpack_encoder_max_dtable_capacity>` = 4096
+ * - :member:`qpack_blocked_streams
+ *   <nghttp3_settings.qpack_blocked_streams>` = 0
+ * - :member:`enable_connect_protocol
+ *   <nghttp3_settings.enable_connect_protocol>` = 0
  */
 NGHTTP3_EXTERN void
 nghttp3_settings_default_versioned(int settings_version,
@@ -2458,7 +2470,7 @@ NGHTTP3_EXTERN uint64_t nghttp3_vec_len(const nghttp3_vec *vec, size_t cnt);
  *
  * `nghttp3_check_header_name` returns nonzero if HTTP header field
  * name |name| of length |len| is valid according to
- * http://tools.ietf.org/html/rfc7230#section-3.2
+ * :rfc:`7230#section-3.2`.
  *
  * Because this is a header field name in HTTP/3, the upper cased
  * alphabet is treated as error.
@@ -2470,7 +2482,7 @@ NGHTTP3_EXTERN int nghttp3_check_header_name(const uint8_t *name, size_t len);
  *
  * `nghttp3_check_header_value` returns nonzero if HTTP header field
  * value |value| of length |len| is valid according to
- * http://tools.ietf.org/html/rfc7230#section-3.2
+ * :rfc:`7230#section-3.2`.
  */
 NGHTTP3_EXTERN int nghttp3_check_header_value(const uint8_t *value, size_t len);
 
