@@ -34,6 +34,10 @@
 #include "nghttp3_conv.h"
 #include "nghttp3_http.h"
 
+/* NGHTTP3_QPACK_ENCODER_MAX_DTABLE_CAPACITY is the upper bound of the
+   dynamic table capacity that QPACK encoder is willing to use. */
+#define NGHTTP3_QPACK_ENCODER_MAX_DTABLE_CAPACITY 4096
+
 /*
  * conn_remote_stream_uni returns nonzero if |stream_id| is remote
  * unidirectional stream ID.
@@ -62,15 +66,16 @@ static int conn_call_begin_headers(nghttp3_conn *conn, nghttp3_stream *stream) {
   return 0;
 }
 
-static int conn_call_end_headers(nghttp3_conn *conn, nghttp3_stream *stream) {
+static int conn_call_end_headers(nghttp3_conn *conn, nghttp3_stream *stream,
+                                 int fin) {
   int rv;
 
   if (!conn->callbacks.end_headers) {
     return 0;
   }
 
-  rv = conn->callbacks.end_headers(conn, stream->node.nid.id, conn->user_data,
-                                   stream->user_data);
+  rv = conn->callbacks.end_headers(conn, stream->node.nid.id, fin,
+                                   conn->user_data, stream->user_data);
   if (rv != 0) {
     /* TODO Allow ignore headers */
     return NGHTTP3_ERR_CALLBACK_FAILURE;
@@ -97,15 +102,16 @@ static int conn_call_begin_trailers(nghttp3_conn *conn,
   return 0;
 }
 
-static int conn_call_end_trailers(nghttp3_conn *conn, nghttp3_stream *stream) {
+static int conn_call_end_trailers(nghttp3_conn *conn, nghttp3_stream *stream,
+                                  int fin) {
   int rv;
 
   if (!conn->callbacks.end_trailers) {
     return 0;
   }
 
-  rv = conn->callbacks.end_trailers(conn, stream->node.nid.id, conn->user_data,
-                                    stream->user_data);
+  rv = conn->callbacks.end_trailers(conn, stream->node.nid.id, fin,
+                                    conn->user_data, stream->user_data);
   if (rv != 0) {
     /* TODO Allow ignore headers */
     return NGHTTP3_ERR_CALLBACK_FAILURE;
@@ -229,13 +235,14 @@ static int conn_new(nghttp3_conn **pconn, int server, int callbacks_version,
   }
 
   rv = nghttp3_qpack_decoder_init(&conn->qdec,
-                                  settings->qpack_max_table_capacity,
+                                  settings->qpack_max_dtable_capacity,
                                   settings->qpack_blocked_streams, mem);
   if (rv != 0) {
     goto qdec_init_fail;
   }
 
-  rv = nghttp3_qpack_encoder_init(&conn->qenc, mem);
+  rv = nghttp3_qpack_encoder_init(
+      &conn->qenc, settings->qpack_encoder_max_dtable_capacity, mem);
   if (rv != 0) {
     goto qenc_init_fail;
   }
@@ -937,12 +944,8 @@ nghttp3_ssize nghttp3_conn_read_control(nghttp3_conn *conn,
 
       /* Fall through */
     case NGHTTP3_CTRL_STREAM_STATE_PRIORITY_UPDATE:
-      /* TODO we need to buffer Priority Field Value because it might
-         be fragmented.  Since our priority header field parser only
-         supports u and i, and does not handle any other structured
-         field items, including parameters, we just buffer minimum
-         length (sizeof("u=D, i") - 1).  If field is longer than that,
-         ignore the frame. */
+      /* We need to buffer Priority Field Value because it might be
+         fragmented. */
       len = (size_t)nghttp3_min(rstate->left, (int64_t)(end - p));
       assert(len > 0);
       if (conn->rx.pri_fieldbuflen == 0 && rstate->left == (int64_t)len) {
@@ -984,18 +987,13 @@ nghttp3_ssize nghttp3_conn_read_control(nghttp3_conn *conn,
 
       if (nghttp3_http_parse_priority(&rstate->fr.priority_update.pri,
                                       pri_field_value,
-                                      pri_field_valuelen) == 0) {
-        /* TODO Draft requires that failing to parse Priority Field
-           Value must be treated as connection error.  Our priority
-           header field parser is quite primitive and does not handle
-           full brown Structured Field Values (we just need to parsing
-           5 bytes here).  In order not to drop connection with an
-           optional data, we ignore PRIORITY_UPDATE if we fail to
-           parse the frame. */
-        rv = nghttp3_conn_on_priority_update(conn, &rstate->fr.priority_update);
-        if (rv != 0) {
-          return rv;
-        }
+                                      pri_field_valuelen) != 0) {
+        return NGHTTP3_ERR_H3_GENERAL_PROTOCOL_ERROR;
+      }
+
+      rv = nghttp3_conn_on_priority_update(conn, &rstate->fr.priority_update);
+      if (rv != 0) {
+        return rv;
       }
 
       conn->rx.pri_fieldbuflen = 0;
@@ -1027,13 +1025,6 @@ static int conn_delete_stream(nghttp3_conn *conn, nghttp3_stream *stream) {
   int bidi = nghttp3_client_stream_bidi(stream->node.nid.id);
   int rv;
 
-  if (bidi && !(stream->flags & NGHTTP3_STREAM_FLAG_SHUT_RD)) {
-    rv = nghttp3_http_on_remote_end_stream(stream);
-    if (rv != 0) {
-      return rv;
-    }
-  }
-
   rv = conn_call_deferred_consume(conn, stream,
                                   nghttp3_stream_get_buffered_datalen(stream));
   if (rv != 0) {
@@ -1405,11 +1396,11 @@ nghttp3_ssize nghttp3_conn_read_bidi(nghttp3_conn *conn, size_t *pnproc,
         }
         /* fall through */
       case NGHTTP3_HTTP_STATE_RESP_HEADERS_BEGIN:
-        rv = conn_call_end_headers(conn, stream);
+        rv = conn_call_end_headers(conn, stream, p == end && fin);
         break;
       case NGHTTP3_HTTP_STATE_REQ_TRAILERS_BEGIN:
       case NGHTTP3_HTTP_STATE_RESP_TRAILERS_BEGIN:
-        rv = conn_call_end_trailers(conn, stream);
+        rv = conn_call_end_trailers(conn, stream, p == end && fin);
         break;
       default:
         /* Unreachable */
@@ -1627,8 +1618,6 @@ int nghttp3_conn_on_settings_entry_received(nghttp3_conn *conn,
                                             const nghttp3_frame_settings *fr) {
   const nghttp3_settings_entry *ent = &fr->iv[0];
   nghttp3_settings *dest = &conn->remote.settings;
-  int rv;
-  size_t max_table_capacity;
 
   /* TODO Check for duplicates */
   switch (ent->id) {
@@ -1636,7 +1625,7 @@ int nghttp3_conn_on_settings_entry_received(nghttp3_conn *conn,
     dest->max_field_section_size = ent->value;
     break;
   case NGHTTP3_SETTINGS_ID_QPACK_MAX_TABLE_CAPACITY:
-    if (dest->qpack_max_table_capacity != 0) {
+    if (dest->qpack_max_dtable_capacity != 0) {
       return NGHTTP3_ERR_H3_SETTINGS_ERROR;
     }
 
@@ -1644,18 +1633,10 @@ int nghttp3_conn_on_settings_entry_received(nghttp3_conn *conn,
       break;
     }
 
-    dest->qpack_max_table_capacity = (size_t)ent->value;
+    dest->qpack_max_dtable_capacity = (size_t)ent->value;
 
-    max_table_capacity = nghttp3_min(4096, ent->value);
-
-    nghttp3_qpack_encoder_set_hard_max_dtable_capacity(&conn->qenc,
-                                                       max_table_capacity);
-
-    rv = nghttp3_qpack_encoder_set_max_dtable_capacity(&conn->qenc,
-                                                       max_table_capacity);
-    if (rv != 0) {
-      return rv;
-    }
+    nghttp3_qpack_encoder_set_max_dtable_capacity(&conn->qenc,
+                                                  (size_t)ent->value);
     break;
   case NGHTTP3_SETTINGS_ID_QPACK_BLOCKED_STREAMS:
     if (dest->qpack_blocked_streams != 0) {
@@ -2518,4 +2499,6 @@ void nghttp3_settings_default_versioned(int settings_version,
 
   memset(settings, 0, sizeof(nghttp3_settings));
   settings->max_field_section_size = NGHTTP3_VARINT_MAX;
+  settings->qpack_encoder_max_dtable_capacity =
+      NGHTTP3_QPACK_ENCODER_MAX_DTABLE_CAPACITY;
 }
