@@ -154,14 +154,10 @@ typedef struct nghttp3_qpack_context {
      NGHTTP3_QPACK_ENTRY_OVERHEAD bytes overhead per each entry. */
   size_t dtable_size;
   size_t dtable_sum;
-  /* hard_max_dtable_capacity is the maximum size of dynamic table.
-     In HTTP/3, it is notified by decoder as
-     SETTINGS_QPACK_MAX_TABLE_CAPACITY.  Any value lower than or equal
-     to SETTINGS_QPACK_MAX_TABLE_CAPACITY is OK because encoder has
-     the authority to decide how many entries are inserted into
-     dynamic table. */
+  /* hard_max_dtable_capacity is the upper bound of
+     max_dtable_capacity. */
   size_t hard_max_dtable_capacity;
-  /* max_dtable_capacity is the effective maximum size of dynamic
+  /* max_dtable_capacity is the maximum capacity of the dynamic
      table. */
   size_t max_dtable_capacity;
   /* max_blocked_streams is the maximum number of stream which can be
@@ -260,8 +256,9 @@ struct nghttp3_qpack_encoder {
 };
 
 /*
- * nghttp3_qpack_encoder_init initializes |encoder|.  |mem| is a
- * memory allocator.
+ * nghttp3_qpack_encoder_init initializes |encoder|.
+ * |hard_max_dtable_capacity| is the upper bound of the dynamic table
+ * capacity.  |mem| is a memory allocator.
  *
  * This function returns 0 if it succeeds, or one of the following
  * negative error codes:
@@ -270,6 +267,7 @@ struct nghttp3_qpack_encoder {
  *     Out of memory.
  */
 int nghttp3_qpack_encoder_init(nghttp3_qpack_encoder *encoder,
+                               size_t hard_max_dtable_capacity,
                                const nghttp3_mem *mem);
 
 /*
@@ -785,9 +783,9 @@ struct nghttp3_qpack_decoder {
 
 /*
  * nghttp3_qpack_decoder_init initializes |decoder|.
- * |max_dtable_capacity| is the maximum size of dynamic table.
- * |max_blocked_streams| is the maximum number of stream which can be
- * blocked.  |mem| is a memory allocator.
+ * |hard_max_dtable_capacity| is the upper bound of the dynamic table
+ * capacity.  |max_blocked_streams| is the maximum number of stream
+ * which can be blocked.  |mem| is a memory allocator.
  *
  * This function returns 0 if it succeeds, or one of the following
  * negative error codes:
@@ -796,7 +794,7 @@ struct nghttp3_qpack_decoder {
  *     Out of memory.
  */
 int nghttp3_qpack_decoder_init(nghttp3_qpack_decoder *decoder,
-                               size_t max_dtable_capacity,
+                               size_t hard_max_dtable_capacity,
                                size_t max_blocked_streams,
                                const nghttp3_mem *mem);
 
