@@ -346,7 +346,7 @@ int nghttp3_stream_write_settings(nghttp3_stream *stream,
   iv[0].id = NGHTTP3_SETTINGS_ID_MAX_FIELD_SECTION_SIZE;
   iv[0].value = local_settings->max_field_section_size;
   iv[1].id = NGHTTP3_SETTINGS_ID_QPACK_MAX_TABLE_CAPACITY;
-  iv[1].value = local_settings->qpack_max_table_capacity;
+  iv[1].value = local_settings->qpack_max_dtable_capacity;
   iv[2].id = NGHTTP3_SETTINGS_ID_QPACK_BLOCKED_STREAMS;
   iv[2].value = local_settings->qpack_blocked_streams;
 
@@ -1055,16 +1055,16 @@ int nghttp3_stream_transit_rx_http_state(nghttp3_stream *stream,
       if (stream->rx.http.flags & NGHTTP3_HTTP_FLAG_METH_CONNECT) {
         return NGHTTP3_ERR_H3_FRAME_UNEXPECTED;
       }
-      rv = nghttp3_http_on_remote_end_stream(stream);
-      if (rv != 0) {
-        return rv;
-      }
       stream->rx.hstate = NGHTTP3_HTTP_STATE_REQ_TRAILERS_BEGIN;
       return 0;
     case NGHTTP3_HTTP_EVENT_DATA_BEGIN:
       stream->rx.hstate = NGHTTP3_HTTP_STATE_REQ_DATA_BEGIN;
       return 0;
     case NGHTTP3_HTTP_EVENT_MSG_END:
+      rv = nghttp3_http_on_remote_end_stream(stream);
+      if (rv != 0) {
+        return rv;
+      }
       stream->rx.hstate = NGHTTP3_HTTP_STATE_REQ_END;
       return 0;
     default:
@@ -1086,13 +1086,13 @@ int nghttp3_stream_transit_rx_http_state(nghttp3_stream *stream,
       if (stream->rx.http.flags & NGHTTP3_HTTP_FLAG_METH_CONNECT) {
         return NGHTTP3_ERR_H3_FRAME_UNEXPECTED;
       }
+      stream->rx.hstate = NGHTTP3_HTTP_STATE_REQ_TRAILERS_BEGIN;
+      return 0;
+    case NGHTTP3_HTTP_EVENT_MSG_END:
       rv = nghttp3_http_on_remote_end_stream(stream);
       if (rv != 0) {
         return rv;
       }
-      stream->rx.hstate = NGHTTP3_HTTP_STATE_REQ_TRAILERS_BEGIN;
-      return 0;
-    case NGHTTP3_HTTP_EVENT_MSG_END:
       stream->rx.hstate = NGHTTP3_HTTP_STATE_REQ_END;
       return 0;
     default:
@@ -1110,6 +1110,10 @@ int nghttp3_stream_transit_rx_http_state(nghttp3_stream *stream,
          spec. */
       return NGHTTP3_ERR_H3_FRAME_UNEXPECTED;
     }
+    rv = nghttp3_http_on_remote_end_stream(stream);
+    if (rv != 0) {
+      return rv;
+    }
     stream->rx.hstate = NGHTTP3_HTTP_STATE_REQ_END;
     return 0;
   case NGHTTP3_HTTP_STATE_REQ_END:
@@ -1137,10 +1141,6 @@ int nghttp3_stream_transit_rx_http_state(nghttp3_stream *stream,
           stream->rx.http.status_code / 100 == 2) {
         return NGHTTP3_ERR_H3_FRAME_UNEXPECTED;
       }
-      rv = nghttp3_http_on_remote_end_stream(stream);
-      if (rv != 0) {
-        return rv;
-      }
       stream->rx.hstate = NGHTTP3_HTTP_STATE_RESP_TRAILERS_BEGIN;
       return 0;
     case NGHTTP3_HTTP_EVENT_DATA_BEGIN:
@@ -1150,6 +1150,10 @@ int nghttp3_stream_transit_rx_http_state(nghttp3_stream *stream,
       stream->rx.hstate = NGHTTP3_HTTP_STATE_RESP_DATA_BEGIN;
       return 0;
     case NGHTTP3_HTTP_EVENT_MSG_END:
+      rv = nghttp3_http_on_remote_end_stream(stream);
+      if (rv != 0) {
+        return rv;
+      }
       stream->rx.hstate = NGHTTP3_HTTP_STATE_RESP_END;
       return 0;
     default:
@@ -1171,13 +1175,13 @@ int nghttp3_stream_transit_rx_http_state(nghttp3_stream *stream,
           stream->rx.http.status_code / 100 == 2) {
         return NGHTTP3_ERR_H3_FRAME_UNEXPECTED;
       }
+      stream->rx.hstate = NGHTTP3_HTTP_STATE_RESP_TRAILERS_BEGIN;
+      return 0;
+    case NGHTTP3_HTTP_EVENT_MSG_END:
       rv = nghttp3_http_on_remote_end_stream(stream);
       if (rv != 0) {
         return rv;
       }
-      stream->rx.hstate = NGHTTP3_HTTP_STATE_RESP_TRAILERS_BEGIN;
-      return 0;
-    case NGHTTP3_HTTP_EVENT_MSG_END:
       stream->rx.hstate = NGHTTP3_HTTP_STATE_RESP_END;
       return 0;
     default:
@@ -1193,6 +1197,10 @@ int nghttp3_stream_transit_rx_http_state(nghttp3_stream *stream,
     if (event != NGHTTP3_HTTP_EVENT_MSG_END) {
       return NGHTTP3_ERR_MALFORMED_HTTP_MESSAGING;
     }
+    rv = nghttp3_http_on_remote_end_stream(stream);
+    if (rv != 0) {
+      return rv;
+    }
     stream->rx.hstate = NGHTTP3_HTTP_STATE_RESP_END;
     return 0;
   case NGHTTP3_HTTP_STATE_RESP_END:
