@@ -103,10 +103,11 @@ static int recv_header(nghttp3_conn *conn, int64_t stream_id, int32_t token,
   return 0;
 }
 
-static int end_headers(nghttp3_conn *conn, int64_t stream_id, void *user_data,
-                       void *stream_user_data) {
+static int end_headers(nghttp3_conn *conn, int64_t stream_id, int fin,
+                       void *user_data, void *stream_user_data) {
   (void)conn;
   (void)stream_id;
+  (void)fin;
   (void)stream_user_data;
   (void)user_data;
   return 0;
@@ -311,7 +312,7 @@ void test_nghttp3_conn_read_control(void) {
 
   CU_ASSERT(nconsumed == (nghttp3_ssize)nghttp3_buf_len(&buf));
   CU_ASSERT(65536 == conn->remote.settings.max_field_section_size);
-  CU_ASSERT(4096 == conn->remote.settings.qpack_max_table_capacity);
+  CU_ASSERT(4096 == conn->remote.settings.qpack_max_dtable_capacity);
   CU_ASSERT(99 == conn->remote.settings.qpack_blocked_streams);
   CU_ASSERT(4096 == conn->qenc.ctx.hard_max_dtable_capacity);
   CU_ASSERT(4096 == conn->qenc.ctx.max_dtable_capacity);
@@ -332,7 +333,7 @@ void test_nghttp3_conn_read_control(void) {
   }
 
   CU_ASSERT(65536 == conn->remote.settings.max_field_section_size);
-  CU_ASSERT(4096 == conn->remote.settings.qpack_max_table_capacity);
+  CU_ASSERT(4096 == conn->remote.settings.qpack_max_dtable_capacity);
   CU_ASSERT(99 == conn->remote.settings.qpack_blocked_streams);
 
   nghttp3_conn_del(conn);
@@ -361,7 +362,7 @@ void test_nghttp3_conn_read_control(void) {
                                        /* fin = */ 0);
 
   CU_ASSERT(nconsumed == (nghttp3_ssize)nghttp3_buf_len(&buf));
-  CU_ASSERT(4097 == conn->remote.settings.qpack_max_table_capacity);
+  CU_ASSERT(4097 == conn->remote.settings.qpack_max_dtable_capacity);
   CU_ASSERT(101 == conn->remote.settings.qpack_blocked_streams);
   CU_ASSERT(4096 == conn->qenc.ctx.hard_max_dtable_capacity);
   CU_ASSERT(4096 == conn->qenc.ctx.max_dtable_capacity);
@@ -641,7 +642,7 @@ void test_nghttp3_conn_http_request(void) {
   callbacks.recv_header = recv_header;
   callbacks.end_headers = end_headers;
 
-  settings.qpack_max_table_capacity = 4096;
+  settings.qpack_max_dtable_capacity = 4096;
   settings.qpack_blocked_streams = 100;
 
   clud.data.left = 2000;
@@ -764,7 +765,7 @@ static void check_http_header(const nghttp3_nv *nva, size_t nvlen, int request,
   settings.enable_connect_protocol = 1;
 
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)nva;
@@ -1113,7 +1114,7 @@ void test_nghttp3_conn_http_content_length(void) {
 
   /* client */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)resnv;
@@ -1137,7 +1138,7 @@ void test_nghttp3_conn_http_content_length(void) {
 
   /* server */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)reqnv;
@@ -1188,9 +1189,31 @@ void test_nghttp3_conn_http_content_length_mismatch(void) {
   memset(&callbacks, 0, sizeof(callbacks));
   nghttp3_settings_default(&settings);
 
-  /* content-length is 20, but no DATA is present */
+  /* content-length is 20, but no DATA is present and see fin */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
+
+  fr.hd.type = NGHTTP3_FRAME_HEADERS;
+  fr.nva = (nghttp3_nv *)reqnv;
+  fr.nvlen = nghttp3_arraylen(reqnv);
+
+  nghttp3_write_frame_qpack(&buf, &qenc, 0, (nghttp3_frame *)&fr);
+
+  nghttp3_conn_server_new(&conn, &callbacks, &settings, mem, NULL);
+  nghttp3_conn_set_max_client_streams_bidi(conn, 1);
+
+  sconsumed = nghttp3_conn_read_stream(conn, 0, buf.pos, nghttp3_buf_len(&buf),
+                                       /* fin = */ 1);
+
+  CU_ASSERT(NGHTTP3_ERR_MALFORMED_HTTP_MESSAGING == sconsumed);
+
+  nghttp3_conn_del(conn);
+  nghttp3_qpack_encoder_free(&qenc);
+
+  /* content-length is 20, but no DATA is present and stream is
+     reset */
+  nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)reqnv;
@@ -1206,16 +1229,20 @@ void test_nghttp3_conn_http_content_length_mismatch(void) {
 
   CU_ASSERT((nghttp3_ssize)nghttp3_buf_len(&buf) == sconsumed);
 
+  rv = nghttp3_conn_shutdown_stream_read(conn, 0);
+
+  CU_ASSERT(0 == rv);
+
   rv = nghttp3_conn_close_stream(conn, 0, NGHTTP3_H3_NO_ERROR);
 
-  CU_ASSERT(NGHTTP3_ERR_MALFORMED_HTTP_MESSAGING == rv);
+  CU_ASSERT(0 == rv);
 
   nghttp3_conn_del(conn);
   nghttp3_qpack_encoder_free(&qenc);
 
   /* content-length is 20, but server receives 21 bytes DATA. */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)reqnv;
@@ -1237,9 +1264,32 @@ void test_nghttp3_conn_http_content_length_mismatch(void) {
 
   /* Check client side as well */
 
-  /* content-length is 20, but no DATA is present */
+  /* content-length is 20, but no DATA is present and see fin */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
+
+  fr.hd.type = NGHTTP3_FRAME_HEADERS;
+  fr.nva = (nghttp3_nv *)resnv;
+  fr.nvlen = nghttp3_arraylen(resnv);
+
+  nghttp3_write_frame_qpack(&buf, &qenc, 0, (nghttp3_frame *)&fr);
+
+  nghttp3_conn_client_new(&conn, &callbacks, &settings, mem, NULL);
+  nghttp3_conn_create_stream(conn, &stream, 0);
+  stream->rx.hstate = NGHTTP3_HTTP_STATE_RESP_INITIAL;
+
+  sconsumed = nghttp3_conn_read_stream(conn, 0, buf.pos, nghttp3_buf_len(&buf),
+                                       /* fin = */ 1);
+
+  CU_ASSERT(NGHTTP3_ERR_MALFORMED_HTTP_MESSAGING == sconsumed);
+
+  nghttp3_conn_del(conn);
+  nghttp3_qpack_encoder_free(&qenc);
+
+  /* content-length is 20, but no DATA is present and stream is
+     reset */
+  nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)resnv;
@@ -1256,16 +1306,20 @@ void test_nghttp3_conn_http_content_length_mismatch(void) {
 
   CU_ASSERT((nghttp3_ssize)nghttp3_buf_len(&buf) == sconsumed);
 
+  rv = nghttp3_conn_shutdown_stream_read(conn, 0);
+
+  CU_ASSERT(0 == rv);
+
   rv = nghttp3_conn_close_stream(conn, 0, NGHTTP3_H3_NO_ERROR);
 
-  CU_ASSERT(NGHTTP3_ERR_MALFORMED_HTTP_MESSAGING == rv);
+  CU_ASSERT(0 == rv);
 
   nghttp3_conn_del(conn);
   nghttp3_qpack_encoder_free(&qenc);
 
   /* content-length is 20, but server receives 21 bytes DATA. */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)resnv;
@@ -1313,7 +1367,7 @@ void test_nghttp3_conn_http_non_final_response(void) {
 
   /* non-final followed by DATA is illegal.  */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)infonv;
@@ -1336,7 +1390,7 @@ void test_nghttp3_conn_http_non_final_response(void) {
 
   /* 2 non-finals followed by final headers */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)infonv;
@@ -1365,7 +1419,7 @@ void test_nghttp3_conn_http_non_final_response(void) {
 
   /* non-finals followed by trailers */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)infonv;
@@ -1425,7 +1479,7 @@ void test_nghttp3_conn_http_trailers(void) {
 
   /* final response followed by trailers */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)resnv;
@@ -1453,7 +1507,7 @@ void test_nghttp3_conn_http_trailers(void) {
 
   /* trailers contain :status */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)resnv;
@@ -1481,7 +1535,7 @@ void test_nghttp3_conn_http_trailers(void) {
 
   /* Receiving 2 trailers HEADERS is invalid*/
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)resnv;
@@ -1511,7 +1565,7 @@ void test_nghttp3_conn_http_trailers(void) {
   /* We don't expect response trailers after HEADERS with CONNECT
      request */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)resnv;
@@ -1542,7 +1596,7 @@ void test_nghttp3_conn_http_trailers(void) {
   /* We don't expect response trailers after DATA with CONNECT
      request */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)resnv;
@@ -1573,7 +1627,7 @@ void test_nghttp3_conn_http_trailers(void) {
 
   /* request followed by trailers */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)reqnv;
@@ -1600,7 +1654,7 @@ void test_nghttp3_conn_http_trailers(void) {
 
   /* request followed by trailers which contains pseudo headers */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)reqnv;
@@ -1622,7 +1676,7 @@ void test_nghttp3_conn_http_trailers(void) {
 
   /* request followed by 2 trailers */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)reqnv;
@@ -1650,7 +1704,7 @@ void test_nghttp3_conn_http_trailers(void) {
 
   /* We don't expect trailers after HEADERS with CONNECT request */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)connect_reqnv;
@@ -1677,7 +1731,7 @@ void test_nghttp3_conn_http_trailers(void) {
 
   /* We don't expect trailers after DATA with CONNECT request */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)connect_reqnv;
@@ -1735,7 +1789,7 @@ void test_nghttp3_conn_http_ignore_content_length(void) {
 
   /* If status code is 304, content-length must be ignored. */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)resnv;
@@ -1762,7 +1816,7 @@ void test_nghttp3_conn_http_ignore_content_length(void) {
 
   /* If method is CONNECT, content-length must be ignored. */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)reqnv;
@@ -1791,7 +1845,7 @@ void test_nghttp3_conn_http_ignore_content_length(void) {
 
   /* Content-Length in 200 response to CONNECT is ignored */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)cl_resnv;
@@ -1850,7 +1904,7 @@ void test_nghttp3_conn_http_record_request_method(void) {
   /* content-length is not allowed with 200 status code in response to
      CONNECT request.  Just ignore it. */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)resnv;
@@ -1876,7 +1930,7 @@ void test_nghttp3_conn_http_record_request_method(void) {
   /* The content-length in response to HEAD request must be
      ignored. */
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.nva = (nghttp3_nv *)resnv;
@@ -1926,7 +1980,7 @@ void test_nghttp3_conn_qpack_blocked_stream(void) {
 
   memset(&callbacks, 0, sizeof(callbacks));
   nghttp3_settings_default(&settings);
-  settings.qpack_max_table_capacity = 4096;
+  settings.qpack_max_dtable_capacity = 4096;
   settings.qpack_blocked_streams = 100;
 
   /* The deletion of QPACK blocked stream is deferred to the moment
@@ -1934,13 +1988,11 @@ void test_nghttp3_conn_qpack_blocked_stream(void) {
   nghttp3_buf_init(&ebuf);
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
 
-  nghttp3_qpack_encoder_init(&qenc, mem);
-  nghttp3_qpack_encoder_set_hard_max_dtable_capacity(
-      &qenc, settings.qpack_max_table_capacity);
+  nghttp3_qpack_encoder_init(&qenc, settings.qpack_max_dtable_capacity, mem);
   nghttp3_qpack_encoder_set_max_blocked_streams(&qenc,
                                                 settings.qpack_blocked_streams);
   nghttp3_qpack_encoder_set_max_dtable_capacity(
-      &qenc, settings.qpack_max_table_capacity);
+      &qenc, settings.qpack_max_dtable_capacity);
 
   nghttp3_conn_client_new(&conn, &callbacks, &settings, mem, NULL);
   nghttp3_conn_bind_qpack_streams(conn, 2, 6);
@@ -1994,13 +2046,11 @@ void test_nghttp3_conn_qpack_blocked_stream(void) {
   nghttp3_buf_init(&ebuf);
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
 
-  nghttp3_qpack_encoder_init(&qenc, mem);
-  nghttp3_qpack_encoder_set_hard_max_dtable_capacity(
-      &qenc, settings.qpack_max_table_capacity);
+  nghttp3_qpack_encoder_init(&qenc, settings.qpack_max_dtable_capacity, mem);
   nghttp3_qpack_encoder_set_max_blocked_streams(&qenc,
                                                 settings.qpack_blocked_streams);
   nghttp3_qpack_encoder_set_max_dtable_capacity(
-      &qenc, settings.qpack_max_table_capacity);
+      &qenc, settings.qpack_max_dtable_capacity);
 
   nghttp3_conn_client_new(&conn, &callbacks, &settings, mem, NULL);
   nghttp3_conn_bind_qpack_streams(conn, 2, 6);
@@ -2440,7 +2490,7 @@ void test_nghttp3_conn_shutdown_server(void) {
   nghttp3_conn_server_new(&conn, &callbacks, &settings, mem, &ud);
   nghttp3_conn_bind_control_stream(conn, 3);
   nghttp3_conn_bind_qpack_streams(conn, 7, 11);
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   fr.hd.type = NGHTTP3_FRAME_HEADERS;
   fr.headers.nva = (nghttp3_nv *)nva;
@@ -2579,7 +2629,7 @@ void test_nghttp3_conn_priority_update(void) {
   nghttp3_conn_bind_control_stream(conn, 3);
   nghttp3_conn_bind_qpack_streams(conn, 7, 11);
   nghttp3_conn_set_max_client_streams_bidi(conn, 1);
-  nghttp3_qpack_encoder_init(&qenc, mem);
+  nghttp3_qpack_encoder_init(&qenc, 0, mem);
 
   buf.last = nghttp3_put_varint(buf.last, NGHTTP3_STREAM_TYPE_CONTROL);
 
@@ -2854,7 +2904,7 @@ void test_nghttp3_conn_shutdown_stream_read(void) {
   memset(&callbacks, 0, sizeof(callbacks));
   callbacks.deferred_consume = deferred_consume;
   nghttp3_settings_default(&settings);
-  settings.qpack_max_table_capacity = 4096;
+  settings.qpack_max_dtable_capacity = 4096;
   settings.qpack_blocked_streams = 100;
 
   /* Shutting down read-side stream when a stream is blocked by QPACK
@@ -2862,13 +2912,11 @@ void test_nghttp3_conn_shutdown_stream_read(void) {
   nghttp3_buf_init(&ebuf);
   nghttp3_buf_wrap_init(&buf, rawbuf, sizeof(rawbuf));
 
-  nghttp3_qpack_encoder_init(&qenc, mem);
-  nghttp3_qpack_encoder_set_hard_max_dtable_capacity(
-      &qenc, settings.qpack_max_table_capacity);
+  nghttp3_qpack_encoder_init(&qenc, settings.qpack_max_dtable_capacity, mem);
   nghttp3_qpack_encoder_set_max_blocked_streams(&qenc,
                                                 settings.qpack_blocked_streams);
   nghttp3_qpack_encoder_set_max_dtable_capacity(
-      &qenc, settings.qpack_max_table_capacity);
+      &qenc, settings.qpack_max_dtable_capacity);
 
   nghttp3_conn_client_new(&conn, &callbacks, &settings, mem, &ud);
   nghttp3_conn_bind_qpack_streams(conn, 2, 6);
