@@ -30,5 +30,7 @@
 #endif /* HAVE_CONFIG_H */
 
 void test_nghttp3_http_parse_priority(void);
+void test_nghttp3_sf_parse_item(void);
+void test_nghttp3_sf_parse_inner_list(void);
 
 #endif /* NGTCP2_CONN_TEST_H */
