@@ -163,11 +163,10 @@ void test_nghttp3_qpack_encoder_encode(void) {
   nghttp3_buf_init(&pbuf);
   nghttp3_buf_init(&rbuf);
   nghttp3_buf_init(&ebuf);
-  rv = nghttp3_qpack_encoder_init(&enc, mem);
+  rv = nghttp3_qpack_encoder_init(&enc, 4096, mem);
 
   CU_ASSERT(0 == rv);
 
-  nghttp3_qpack_encoder_set_hard_max_dtable_capacity(&enc, 4096);
   nghttp3_qpack_encoder_set_max_blocked_streams(&enc, 1);
 
   nghttp3_qpack_encoder_set_max_dtable_capacity(&enc, 4096);
@@ -250,11 +249,10 @@ void test_nghttp3_qpack_encoder_still_blocked(void) {
   nghttp3_buf_init(&pbuf);
   nghttp3_buf_init(&rbuf);
   nghttp3_buf_init(&ebuf);
-  rv = nghttp3_qpack_encoder_init(&enc, mem);
+  rv = nghttp3_qpack_encoder_init(&enc, 4096, mem);
 
   CU_ASSERT(0 == rv);
 
-  nghttp3_qpack_encoder_set_hard_max_dtable_capacity(&enc, 4096);
   nghttp3_qpack_encoder_set_max_blocked_streams(&enc, 1);
 
   nghttp3_qpack_encoder_set_max_dtable_capacity(&enc, 4096);
@@ -327,11 +325,10 @@ void test_nghttp3_qpack_encoder_set_dtable_cap(void) {
   nghttp3_buf_init(&rbuf);
   nghttp3_buf_init(&ebuf);
 
-  rv = nghttp3_qpack_encoder_init(&enc, mem);
+  rv = nghttp3_qpack_encoder_init(&enc, 4096, mem);
 
   CU_ASSERT(0 == rv);
 
-  nghttp3_qpack_encoder_set_hard_max_dtable_capacity(&enc, 4096);
   nghttp3_qpack_encoder_set_max_blocked_streams(&enc, 3);
 
   nghttp3_qpack_encoder_set_max_dtable_capacity(&enc, 4096);
@@ -389,9 +386,8 @@ void test_nghttp3_qpack_encoder_set_dtable_cap(void) {
   nghttp3_buf_reset(&rbuf);
   nghttp3_buf_reset(&ebuf);
 
-  rv = nghttp3_qpack_encoder_set_max_dtable_capacity(&enc, 0);
+  nghttp3_qpack_encoder_set_max_dtable_capacity(&enc, 0);
 
-  CU_ASSERT(0 == rv);
   CU_ASSERT(0 == enc.ctx.max_dtable_capacity);
   CU_ASSERT(2 == nghttp3_qpack_encoder_get_num_blocked_streams(&enc));
 
@@ -481,8 +477,7 @@ void test_nghttp3_qpack_encoder_set_dtable_cap(void) {
   nghttp3_qpack_encoder_free(&enc);
 
   /* Check that minimum size is emitted */
-  nghttp3_qpack_encoder_init(&enc, mem);
-  nghttp3_qpack_encoder_set_hard_max_dtable_capacity(&enc, 4096);
+  nghttp3_qpack_encoder_init(&enc, 4096, mem);
   nghttp3_qpack_encoder_set_max_blocked_streams(&enc, 1);
   nghttp3_qpack_encoder_set_max_dtable_capacity(&enc, 4096);
   nghttp3_qpack_decoder_init(&dec, 4096, 1, mem);
@@ -509,13 +504,9 @@ void test_nghttp3_qpack_encoder_set_dtable_cap(void) {
   nghttp3_buf_reset(&rbuf);
   nghttp3_buf_reset(&ebuf);
 
-  rv = nghttp3_qpack_encoder_set_max_dtable_capacity(&enc, 0);
+  nghttp3_qpack_encoder_set_max_dtable_capacity(&enc, 0);
+  nghttp3_qpack_encoder_set_max_dtable_capacity(&enc, 1024);
 
-  CU_ASSERT(0 == rv);
-
-  rv = nghttp3_qpack_encoder_set_max_dtable_capacity(&enc, 1024);
-
-  CU_ASSERT(0 == rv);
   CU_ASSERT(0 == enc.min_dtable_update);
   CU_ASSERT(1024 == enc.last_max_dtable_update);
 
@@ -584,11 +575,10 @@ void test_nghttp3_qpack_decoder_feedback(void) {
 
   nghttp3_buf_reserve(&dbuf, 4096, mem);
 
-  rv = nghttp3_qpack_encoder_init(&enc, mem);
+  rv = nghttp3_qpack_encoder_init(&enc, 4096, mem);
 
   CU_ASSERT(0 == rv);
 
-  nghttp3_qpack_encoder_set_hard_max_dtable_capacity(&enc, 4096);
   nghttp3_qpack_encoder_set_max_blocked_streams(&enc, 2);
 
   nghttp3_qpack_encoder_set_max_dtable_capacity(&enc, 4096);
